import io
import zlib
import struct
from math import ceil

from libc.stdlib cimport abort
from libc.string cimport memset
from cython.parallel import prange

from czlib cimport *
from cpython_nogil cimport *


cdef enum:
    BLOCK_BATCH_SIZE = 300
    MAGIC_LENGTH = 4

block_batch_size = int(BLOCK_BATCH_SIZE)  # make BLOCK_BATCH_SIZE accessible in Python

cdef enum bgzip_err:
    BGZIP_CRC_MISMATCH = -8
    BGZIP_ZLIB_INITIALIZATION_ERROR
    BGZIP_BLOCK_SIZE_MISMATCH
    BGZIP_BLOCK_SIZE_NEGATIVE
    BGZIP_ZLIB_ERROR
    BGZIP_MALFORMED_HEADER
    BGZIP_INSUFFICIENT_BYTES
    BGZIP_ERROR
    BGZIP_OK

cdef const unsigned char * MAGIC = "\037\213\010\4"
# cdef Bytef * HEADER = b"\037\213\010\4\0\0\0\0\0\377\6\0\102\103\2\0"
cdef bytes HEADER = b"\037\213\010\4\0\0\0\0\0\377\6\0\102\103\2\0"

ctypedef block_header_s BlockHeader
cdef struct block_header_s:
    unsigned char magic[MAGIC_LENGTH]
    unsigned int mod_time
    unsigned char extra_flags
    unsigned char os_type
    unsigned short extra_len

ctypedef block_header_subfield_s BlockHeaderSubfield
cdef struct block_header_subfield_s:
    unsigned char id_[2]
    unsigned short length

ctypedef block_header_bgzip_subfield_s BlockHeaderBGZipSubfield
cdef struct block_header_bgzip_subfield_s:
    unsigned char id_[2]
    unsigned short length
    unsigned short block_size

ctypedef block_tailer_s BlockTailer
cdef struct block_tailer_s:
    unsigned int crc
    unsigned int inflated_size

ctypedef block_s Block
cdef struct block_s:
    unsigned int deflated_size
    unsigned int inflated_size
    unsigned int crc
    unsigned short block_size
    Bytef * next_in
    unsigned int available_in
    Bytef * next_out
    unsigned int avail_out

ctypedef bgzip_stream_s BGZipStream
cdef struct bgzip_stream_s:
    unsigned int available_in
    Bytef *next_in

ctypedef chunk_s Chunk
cdef struct chunk_s:
    BGZipStream src
    Block * blocks
    unsigned int num_blocks
    unsigned int inflated_size
    unsigned int bytes_read

class BGZIPException(Exception):
    pass

class BGZIPMalformedHeaderException(BGZIPException):
    pass

cdef bgzip_err inflate_block(Block * block) nogil:
    cdef z_stream zst
    cdef int err

    zst.zalloc = NULL
    zst.zfree = NULL
    zst.opaque = NULL
    zst.avail_in = block.deflated_size
    zst.avail_out = 1024 * 1024
    zst.next_in = block.next_in
    zst.next_out = block.next_out

    err = inflateInit2(&zst, -15)
    if Z_OK != err:
        return BGZIP_ZLIB_INITIALIZATION_ERROR
    err = inflate(&zst, Z_FINISH)
    if Z_STREAM_END == err:
        pass
    else:
        return BGZIP_ZLIB_ERROR
    inflateEnd(&zst)

    if block[0].inflated_size != zst.total_out:
        return BGZIP_BLOCK_SIZE_MISMATCH

    if block.crc != crc32(0, block.next_out, block.inflated_size):
        return BGZIP_CRC_MISMATCH

    return BGZIP_OK

    # Difference betwwen `compress` and `deflate`:
    # https://stackoverflow.com/questions/10166122/zlib-differences-between-the-deflate-and-compress-functions

cdef void * ref_and_advance(BGZipStream * rb, unsigned int member_size, bgzip_err *err) nogil:
    if rb.available_in  < member_size:
        err[0] = BGZIP_INSUFFICIENT_BYTES
        return NULL
    else:
        ret_val = rb.next_in
        rb.next_in += member_size
        rb.available_in  -= member_size
        err[0] = BGZIP_OK
        return ret_val

cdef bgzip_err read_block(Block * block, BGZipStream *src) nogil:
    cdef unsigned int i
    cdef bgzip_err err
    cdef BlockHeader * head
    cdef BlockTailer * tail
    cdef BlockHeaderSubfield * subfield
    cdef Bytef * subfield_data
    cdef unsigned int extra_len

    block.block_size = 0

    head = <BlockHeader *>ref_and_advance(src, sizeof(BlockHeader), &err)
    if err:
        return err

    for i in range(<unsigned int>MAGIC_LENGTH):
        if head.magic[i] != MAGIC[i]:
            return BGZIP_MALFORMED_HEADER

    extra_len = head.extra_len
    while extra_len > 0:
        subfield = <BlockHeaderSubfield *>ref_and_advance(src, sizeof(BlockHeaderSubfield), &err)
        if err:
            return err
        extra_len -= sizeof(BlockHeaderSubfield)

        subfield_data = <Bytef *>ref_and_advance(src, subfield.length, &err)
        if err:
            return err
        extra_len -= sizeof(subfield.length)

        if b"B" == subfield.id_[0] and b"C" == subfield.id_[1]:
            if subfield.length != 2:
                return BGZIP_MALFORMED_HEADER
            block.block_size = (<unsigned short *>subfield_data)[0]

    if 0 != extra_len:
        return BGZIP_BLOCK_SIZE_MISMATCH

    if 0 >= block.block_size:
        return BGZIP_BLOCK_SIZE_NEGATIVE

    block.next_in = src.next_in
    block.deflated_size = 1 + block.block_size - sizeof(BlockHeader) - head.extra_len - sizeof(BlockTailer)

    ref_and_advance(src, block.deflated_size, &err)
    if err:
        return err

    tail = <BlockTailer *>ref_and_advance(src, sizeof(BlockTailer), &err)
    if err:
        return err

    block.crc = tail.crc
    block.inflated_size = tail.inflated_size

cdef py_memoryview_to_buffer(object py_memoryview, Bytef ** buf):
    cdef PyObject * obj = <PyObject *>py_memoryview
    if PyMemoryView_Check(obj):
        # TODO: Check buffer is contiguous, has normal stride
        buf[0] = <Bytef *>(<Py_buffer *>PyMemoryView_GET_BUFFER(obj)).buf
        assert NULL != buf
    else:
        raise TypeError("'py_memoryview' must be a memoryview instance.")

cdef void read_chunk(Chunk *chunk, int blocks_available, unsigned int output_bytes_available) nogil:
    cdef int i = 0
    cdef BGZipStream curr

    for i in range(blocks_available):
        if not chunk[0].src.available_in:
            break
        curr = chunk[0].src
        err = read_block(&chunk[0].blocks[i], &chunk[0].src)
        if BGZIP_OK == err:
            pass
        elif BGZIP_INSUFFICIENT_BYTES == err:
            chunk[0].src = curr
            break
        elif BGZIP_MALFORMED_HEADER == err:
            raise BGZIPMalformedHeaderException("Block gzip magic not found in header.")
        else:
            raise BGZIPException("decompress 2 error")
        if output_bytes_available < chunk[0].inflated_size + chunk[0].blocks[i].inflated_size:
            chunk[0].src = curr
            break
        chunk[0].num_blocks += 1
        chunk[0].inflated_size += chunk[0].blocks[i].inflated_size
        chunk[0].bytes_read += 1 + chunk[0].blocks[i].block_size

def inflate_chunks(list py_chunks, object py_dst_buf, int num_threads, atomic: bool=False):
    """
    Inflate bytes from `py_chunks` into `dst_buff`
    """
    cdef int i, err, num_chunks_read, num_src_chunks = 0, num_blocks_read = 0, _atomic = int(atomic)
    cdef Bytef * dst_buf = NULL
    cdef Block blocks[BLOCK_BATCH_SIZE]
    cdef Chunk atom, chunks[BLOCK_BATCH_SIZE]

    memset(&chunks[0], 0, BLOCK_BATCH_SIZE * sizeof(Chunk))

    num_src_chunks = min(len(py_chunks), BLOCK_BATCH_SIZE)
    for i in range(num_src_chunks):
        py_memoryview_to_buffer(py_chunks[i], &chunks[i].src.next_in)
        chunks[i].src.available_in = len(py_chunks[i])

    py_memoryview_to_buffer(py_dst_buf, &dst_buf)
    cdef unsigned int avail_out = PySequence_Size(<PyObject *>py_dst_buf)

    with nogil:
        for i in range(num_src_chunks):
            chunks[i].blocks = &blocks[num_blocks_read]
            atom = chunks[i]
            read_chunk(&chunks[i], BLOCK_BATCH_SIZE - num_blocks_read, avail_out)
            avail_out -= chunks[i].inflated_size
            num_blocks_read += chunks[i].num_blocks
            if chunks[i].src.available_in:
                if _atomic:
                    num_blocks_read -= chunks[i].num_blocks
                    chunks[i] = atom
                    i -= 1
                break

        num_chunks_read = 1 + i if num_blocks_read else 0

        for i in range(num_blocks_read):
            blocks[i].next_out = dst_buf
            dst_buf += blocks[i].inflated_size

        for i in prange(num_blocks_read, num_threads=num_threads, schedule="dynamic"):
            inflate_block(&blocks[i])

    remaining_chunks = list()
    for i, py_chunk in enumerate(py_chunks):
        if i < BLOCK_BATCH_SIZE:
            if chunks[i].bytes_read < len(py_chunk):
                remaining_chunks.append(py_chunk[chunks[i].bytes_read:])
        else:
            remaining_chunks.append(py_chunk)

    return {'bytes_read':       sum(chunks[i].bytes_read for i in range(num_chunks_read)),
            'bytes_inflated':   sum(chunks[i].inflated_size for i in range(num_chunks_read)),
            'remaining_chunks': remaining_chunks,
            'block_sizes':      [blocks[i].inflated_size for i in range(num_blocks_read)],
            'blocks_per_chunk': [chunks[i].num_blocks for i in range(num_chunks_read)]}

cdef bgzip_err compress_block(Block * block) nogil:
    cdef z_stream zst
    cdef int err = 0
    cdef BlockHeader * head
    cdef BlockHeaderBGZipSubfield * head_subfield
    cdef BlockTailer * tail
    cdef int wbits = -15
    cdef int mem_level = 8

    head = <BlockHeader *>block.next_out
    block.next_out += sizeof(BlockHeader)

    head_subfield = <BlockHeaderBGZipSubfield *>block.next_out
    block.next_out += sizeof(BlockHeaderBGZipSubfield)

    zst.zalloc = NULL
    zst.zfree = NULL
    zst.opaque = NULL
    zst.next_in = block.next_in
    zst.avail_in = block.available_in
    zst.next_out = block.next_out
    zst.avail_out = 1024 * 1024
    err = deflateInit2(&zst, Z_BEST_COMPRESSION, Z_DEFLATED, wbits, mem_level, Z_DEFAULT_STRATEGY)
    if Z_OK != err:
        return BGZIP_ZLIB_ERROR
    err = deflate(&zst, Z_FINISH)
    if Z_STREAM_END != err:
        return BGZIP_ZLIB_ERROR
    deflateEnd(&zst)

    block.next_out += zst.total_out

    tail = <BlockTailer *>block.next_out

    for i in range(MAGIC_LENGTH):
        head.magic[i] = MAGIC[i]
    head.mod_time = 0
    head.extra_flags = 0
    head.os_type = b"\377"
    head.extra_len = sizeof(BlockHeaderBGZipSubfield)

    head_subfield.id_[0] = b"B"
    head_subfield.id_[1] = b"C"
    head_subfield.length = 2
    head_subfield.block_size = sizeof(BlockHeader) + sizeof(BlockHeaderBGZipSubfield) + zst.total_out + sizeof(BlockTailer) - 1

    tail.crc = crc32(0, block.next_in, block.inflated_size)
    tail.inflated_size = block.inflated_size

    block.block_size = 1 + head_subfield.block_size

    return BGZIP_OK

cdef unsigned int _block_data_inflated_size = 65280
cdef unsigned int _block_metadata_size = sizeof(BlockHeader) + sizeof(BlockHeaderBGZipSubfield) + sizeof(BlockTailer)
block_data_inflated_size = _block_data_inflated_size
block_metadata_size = _block_metadata_size

cdef void _get_buffer(PyObject * obj, Py_buffer * view):
    cdef int err

    err = PyObject_GetBuffer(obj, view, PyBUF_SIMPLE)
    if -1 == err:
        raise Exception()

def deflate_to_buffers(py_input_buff, list py_deflated_buffers, int num_threads):
    """
    Compress the data in `py_input_buff` and write it to `handle`.

    `deflated_buffers` should contain enough buffers to hold the number of blocks compressed. Each
    buffer should hold `_block_data_inflated_size + _block_metadata_size` bytes.
    """
    cdef int i, chunk_size
    cdef unsigned int bytes_available = len(py_input_buff)
    cdef int number_of_chunks = min(ceil(bytes_available / block_data_inflated_size),
                                    len(py_deflated_buffers))
    cdef Block blocks[BLOCK_BATCH_SIZE]
    cdef PyObject * deflated_buffers = <PyObject *>py_deflated_buffers
    cdef PyObject * compressed_chunk

    cdef Py_buffer input_view 
    _get_buffer(<PyObject *>py_input_buff, &input_view)

    with nogil:
        for i in range(number_of_chunks):
            compressed_chunk = <PyObject *>PyList_GetItem(deflated_buffers, i)

            if bytes_available >= _block_data_inflated_size:
                chunk_size = _block_data_inflated_size
            else:
                chunk_size = bytes_available

            bytes_available -= _block_data_inflated_size

            blocks[i].inflated_size = chunk_size
            blocks[i].next_in = <Bytef *>input_view.buf + (i * _block_data_inflated_size)
            blocks[i].available_in = chunk_size
            blocks[i].next_out = <Bytef *>PyByteArray_AS_STRING(compressed_chunk)
            blocks[i].avail_out = _block_data_inflated_size + _block_metadata_size

        for i in prange(number_of_chunks, num_threads=num_threads, schedule="dynamic"):
            if BGZIP_OK != compress_block(&blocks[i]):
                with gil:
                    raise BGZIPException()

    PyBuffer_Release(&input_view)

    return [blocks[i].block_size for i in range(number_of_chunks)]



ctypedef unsigned char Bytef
ctypedef unsigned long uLongf
ctypedef unsigned int uInt
ctypedef z_stream_s z_stream

cdef extern from "zlib.h":
    extern struct z_stream_s:
        void * zalloc(void *, unsigned, unsigned) nogil
        void zfree(void *, void *) nogil
        uInt avail_in
        uInt avail_out
        const Bytef *next_in
        const Bytef *next_out
        uLongf total_out
        void * opaque

    extern int inflate(z_stream * strm, int flush) nogil
    extern int inflateInit2(z_stream * strm, int wbits) nogil
    extern int inflateEnd(z_stream *) nogil

    extern int deflate(z_stream * strm, int flush) nogil
    extern int deflateInit2(z_stream * strm, int level, int method, int wbits, int  mem_level, int strategy) nogil
    extern int deflateEnd(z_stream *) nogil

    extern uLongf crc32(uLongf crc, const Bytef * data, unsigned int len) nogil

    extern int Z_OK
    extern int Z_FINISH
    extern int Z_STREAM_END
    extern int Z_BEST_COMPRESSION
    extern int Z_DEFLATED
    extern int Z_DEFAULT_STRATEGY



#!/usr/bin/env python
import io
import os
import sys
import gzip
import random
import unittest
from random import randint
from typing import Any, Generator, List, Sequence

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

import bgzip


class TestBGZipReader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
            with gzip.GzipFile(fileobj=raw) as fh:
                cls.expected_data = fh.read()

    def test_read(self):
        with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
            with bgzip.BGZipReader(raw, 1024 * 1024 * 1) as fh:
                data = bytearray()
                while True:
                    d = fh.read(randint(1024 * 1, 1024 * 1024 * 1024))
                    if not d:
                        break
                    data.extend(d)
                    d.release()
        self.assertEqual(self.expected_data, data)

    def test_empty(self):
        with bgzip.BGZipReader(io.BytesIO()) as fh:
            d = fh.read(1024)
            self.assertEqual(0, len(d))

    def test_read_all(self):
        with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
            with bgzip.BGZipReader(raw) as fh:
                data = fh.read()
        self.assertEqual(data, self.expected_data)

    def test_read_into(self):
        with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
            data = bytearray()
            with bgzip.BGZipReader(raw) as fh:
                while True:
                    d = fh.read(30 * 1024 * 1024)
                    if not d:
                        break
                    data.extend(d)
                    d.release()
        self.assertEqual(self.expected_data, data)

    def test_iter(self):
        with self.subTest("iter byte lines"):
            data = b""
            with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
                with bgzip.BGZipReader(raw, 1024 * 1024 * 1) as fh:
                    for line in fh:
                        data += line
            self.assertEqual(self.expected_data, data)

        with self.subTest("iter text lines"):
            content = ""
            with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
                with bgzip.BGZipReader(raw, 1024 * 1024 * 1) as fh:
                    with io.TextIOWrapper(fh, "utf-8") as handle:
                        for line in handle:
                            content += line
            self.assertEqual(self.expected_data.decode("utf-8"), content)

    def test_inflate_chunks(self):
        size = (2 * bgzip.bgu.block_batch_size + 1) * bgzip.bgu.block_data_inflated_size
        expected_data = os.urandom(size)
        inflate_buf = memoryview(bytearray(30 * 1024 * 1024))

        data = memoryview(expected_data)
        deflated_blocks = list()
        deflater = bgzip.Deflater()
        while data:
            bytes_deflated, blocks = deflater.deflate(data)
            deflated_blocks.extend([bytes(b) for b in blocks])
            data = data[bytes_deflated:]

        def _test_inflate_chunks(remaining_chunks: List[memoryview], atomic: bool=False):
            remaining_chunks = remaining_chunks.copy()
            reinflated_data = b""
            while remaining_chunks:
                inflate_info = bgzip.inflate_chunks(remaining_chunks, inflate_buf, atomic=atomic)
                self.assertGreater(inflate_info['bytes_inflated'], 0)
                remaining_chunks = inflate_info['remaining_chunks']
                reinflated_data += b"".join(inflate_info['blocks'])
            self.assertEqual(expected_data, reinflated_data)

        with self.subTest("all blocks"):
            _test_inflate_chunks([memoryview(bytes(b)) for b in deflated_blocks])

        with self.subTest("chunked blocks"):
            _test_inflate_chunks([memoryview(b"".join(chunk))
                                  for chunk in _randomly_chunked(deflated_blocks)])

        with self.subTest("initial large chunk"):
            _test_inflate_chunks([memoryview(b"".join(deflated_blocks[:-1])),
                                  memoryview(deflated_blocks[-1])])

        with self.subTest("trailing large chunk"):
            _test_inflate_chunks([memoryview(deflated_blocks[0]),
                                  memoryview(b"".join(deflated_blocks[1:]))])

        with self.subTest("leading large chunk atomic"):
            inflate_buf = memoryview(bytearray(200 * 1024))
            chunks = [memoryview(b"".join(deflated_blocks[:-1])), memoryview(deflated_blocks[-1])]
            inflate_info = bgzip.inflate_chunks(chunks, inflate_buf, atomic=True)
            self.assertEqual(inflate_info['remaining_chunks'], chunks)

        with self.subTest("trailing large chunk atomic"):
            inflate_buf = memoryview(bytearray(200 * 1024))
            chunks = [memoryview(deflated_blocks[0]), memoryview(b"".join(deflated_blocks[1:]))]
            inflate_info = bgzip.inflate_chunks(chunks, inflate_buf, atomic=True)
            self.assertEqual(inflate_info['remaining_chunks'][0], chunks[1])

        with self.subTest("small inflate buf"):
            inflate_buf = memoryview(bytearray(200 * 1024))
            _test_inflate_chunks([memoryview(b"".join(chunk))
                                  for chunk in _randomly_chunked(deflated_blocks)])

        with self.subTest("buf too small to inflate anything"):
            inflate_buf = memoryview(bytearray(1))
            chunks = [memoryview(b"".join(deflated_blocks))]
            inflate_info = bgzip.inflate_chunks(chunks, inflate_buf)
            self.assertEqual(0, inflate_info['bytes_read'])
            self.assertEqual(0, inflate_info['bytes_inflated'])
            self.assertEqual(chunks, inflate_info['remaining_chunks'])
            self.assertEqual(list(), inflate_info['block_sizes'])
            self.assertEqual(list(), inflate_info['blocks_per_chunk'])
            self.assertEqual(list(), inflate_info['blocks'])

        with self.subTest("passing in non-memoryview buffers should raise"):
            with self.assertRaises(TypeError):
                bgzip.inflate_chunks([b"asfd"], inflate_buf)

    def test_inflate_streamed_chunk(self):
        with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
            chunk = raw.read()
        with gzip.GzipFile(fileobj=io.BytesIO(chunk)) as fh:
            expected_data = fh.read()
        inflate_buf = memoryview(bytearray(1024 * 1024 * 50))
        input_buf, data = bytes(), bytes()
        with open("tests/fixtures/partial.vcf.gz", "rb") as raw:
            while True:
                input_buf += raw.read(random.randint(0, 100 * 1024))
                if not input_buf:
                    break
                inflate_info = bgzip.inflate_chunks([memoryview(input_buf)], inflate_buf)
                input_buf = b"".join(inflate_info['remaining_chunks'])
                data += b"".join(inflate_info['blocks'])
        self.assertEqual(expected_data, data)

def _randomly_chunked(items: Sequence[Any]) -> Generator[Sequence[Any], None, None]:
    items = [i for i in items]
    while items:
        chunk_size = random.randint(1, 20)
        yield items[:chunk_size]
        items = items[chunk_size:]

class TestBGZipWriter(unittest.TestCase):
    def test_gen_buffers(self):
        bgzip.Deflater._gen_buffers(bgzip.bgu.block_batch_size)
        bgzip.Deflater._gen_buffers(1)

        for num_bufs in [0, bgzip.bgu.block_batch_size + 1]:
            with self.assertRaises(ValueError):
                bgzip.Deflater._gen_buffers(num_bufs)

    def test_write(self):
        inflated_data = os.urandom(1024 * 1024 * 50)
        deflater = bgzip.Deflater()
        deflated_with_buffers = bytes()
        data = memoryview(bytes(inflated_data))
        while data:
            bytes_deflated, deflated_blocks = deflater.deflate(data)
            data = data[bytes_deflated:]
            deflated_with_buffers += b"".join(deflated_blocks)
        deflated_with_buffers += bgzip.bgzip_eof

        fh_out = io.BytesIO()
        with bgzip.BGZipWriter(fh_out) as writer:
            n = 987345
            writer.write(inflated_data[:n])
            writer.write(inflated_data[n:])
        deflated_with_writer = fh_out.getvalue()

        self.assertEqual(deflated_with_buffers, deflated_with_writer)

        fh_out.seek(0)
        with gzip.GzipFile(fileobj=fh_out) as fh:
            reinflated_data = fh.read()

        self.assertEqual(inflated_data, reinflated_data)
        self.assertTrue(deflated_with_writer.endswith(bgzip.bgzip_eof))

    def test_write_random_data(self):
        inflated_data = os.urandom(1024 * 1024)
        with bgzip.BGZipWriter(io.BytesIO()) as writer:
            writer.write(inflated_data)

    def test_pathalogical_write(self):
        fh = io.BytesIO()
        with bgzip.BGZipWriter(fh):
            fh.write(b"")

    def test_large_write(self):
        """Force write to use several batch calls to bgzip_utils."""
        fh_out = io.BytesIO()
        with bgzip.BGZipWriter(fh_out) as writer:
            number_of_blocks = 2 * bgzip.bgu.block_batch_size + 1
            size = number_of_blocks * bgzip.bgu.block_data_inflated_size
            writer.write(bytearray(size))

if __name__ == '__main__':
    unittest.main()



import io
from math import floor, ceil
from multiprocessing import cpu_count
from typing import Any, Dict, Generator, IO, List, Sequence, Tuple

from bgzip import bgzip_utils as bgu  # type: ignore


# samtools format specs:
# https://samtools.github.io/hts-specs/SAMv1.pdf
bgzip_eof = bytes.fromhex("1f8b08040000000000ff0600424302001b0003000000000000000000")

DEFAULT_DECOMPRESS_BUFFER_SZ = 1024 * 1024 * 50

class BGZipReader(io.RawIOBase):
    """
    Inflate data into a pre-allocated buffer. The buffer size will not change, and should be large enough
    to hold at least twice the data of any call to `read`.
    """
    def __init__(self,
                 fileobj: IO,
                 buffer_size: int=DEFAULT_DECOMPRESS_BUFFER_SZ,
                 num_threads=cpu_count(),
                 raw_read_chunk_size=256 * 1024):
        self.fileobj = fileobj
        self._input_data = bytes()
        self._inflate_buf = memoryview(bytearray(buffer_size))
        self._start = self._stop = 0
        self.raw_read_chunk_size = raw_read_chunk_size
        self.num_threads = num_threads

    def readable(self) -> bool:
        return True

    def _fetch_and_inflate(self):
        while True:
            self._input_data += self.fileobj.read(self.raw_read_chunk_size)
            inflate_info = bgu.inflate_chunks([memoryview(self._input_data)],
                                              self._inflate_buf[self._start:],
                                              num_threads=self.num_threads)
            if self._input_data and not inflate_info['bytes_inflated']:
                # Not enough space at end of buffer, reset indices
                assert self._start == self._stop, "Read error. Please contact bgzip maintainers."
                self._start = self._stop = 0
            else:
                self._input_data = self._input_data[inflate_info['bytes_read']:]
                self._stop += inflate_info['bytes_inflated']
                break

    def _read(self, requested_size: int) -> memoryview:
        if self._start == self._stop:
            self._fetch_and_inflate()
        size = min(requested_size, self._stop - self._start)
        out = self._inflate_buf[self._start:self._start + size]
        self._start += len(out)
        return out

    def read(self, size: int=-1) -> memoryview:
        """
        Return a view to mutable memory. View should be consumed before calling 'read' again.
        """
        if -1 == size:
            data = bytearray()
            while True:
                try:
                    d = self._read(1024 ** 3)
                    if not d:
                        break
                    data.extend(d)
                finally:
                    d.release()
            out = memoryview(data)
        else:
            out = self._read(size)
        return out

    def readinto(self, buff) -> int:
        sz = len(buff)
        bytes_read = 0
        while bytes_read < sz:
            mv = self.read(sz - bytes_read)
            if not mv:
                break
            buff[bytes_read:bytes_read + len(mv)] = mv
            bytes_read += len(mv)
        return bytes_read

    def __iter__(self) -> Generator[bytes, None, None]:
        if not hasattr(self, "_buffered"):
            self._buffered = io.BufferedReader(self)
        for line in self._buffered:
            yield line

    def close(self):
        super().close()
        if hasattr(self, "_buffered"):
            self._buffered.close()

def inflate_chunks(chunks: Sequence[memoryview],
                   inflate_buf: memoryview,
                   num_threads: int=cpu_count(),
                   atomic: bool=False) -> Dict[str, Any]:
    inflate_info = bgu.inflate_chunks(chunks, inflate_buf, num_threads, atomic=atomic)
    blocks: List[memoryview] = [None] * len(inflate_info['block_sizes'])  # type: ignore
    total = 0
    for i, sz in enumerate(inflate_info['block_sizes']):
        blocks[i] = inflate_buf[total: total + sz]
        total += sz
    inflate_info['blocks'] = blocks
    return inflate_info

class BGZipWriter(io.IOBase):
    def __init__(self, fileobj: IO, num_threads: int=cpu_count()):
        self.fileobj = fileobj
        self._input_buffer = bytearray()
        self._deflater = Deflater(num_threads)
        self.num_threads = num_threads

    def writable(self):
        return True

    def _compress(self, process_all_chunks=False):
        while self._input_buffer:
            bytes_deflated, blocks = self._deflater.deflate(self._input_buffer)
            for b in blocks:
                self.fileobj.write(b)
            self._input_buffer = self._input_buffer[bytes_deflated:]
            if len(self._input_buffer) < bgu.block_data_inflated_size and not process_all_chunks:
                break

    def write(self, data):
        self._input_buffer.extend(data)
        if len(self._input_buffer) > bgu.block_batch_size * bgu.block_data_inflated_size:
            self._compress()

    def close(self):
        if self._input_buffer:
            self._compress(process_all_chunks=True)
        self.fileobj.write(bgzip_eof)
        self.fileobj.flush()

class Deflater:
    def __init__(self, num_threads: int=cpu_count(), num_deflate_buffers: int=bgu.block_batch_size):
        self._num_threads = num_threads
        self._deflate_bufs = self._gen_buffers(num_deflate_buffers)

    @staticmethod
    def _gen_buffers(number_of_buffers: int=bgu.block_batch_size) -> List[bytearray]:
        if 0 >= number_of_buffers or bgu.block_batch_size < number_of_buffers:
            raise ValueError(f"0 < 'number_of_buffers' <= '{bgu.block_batch_size}")
        # Include a kilobyte of padding for poorly compressible data
        max_deflated_block_size = bgu.block_data_inflated_size + bgu.block_metadata_size + 1024
        return [bytearray(max_deflated_block_size) for _ in range(number_of_buffers)]

    def deflate(self, data: memoryview) -> Tuple[int, List[memoryview]]:
        deflated_sizes = bgu.deflate_to_buffers(data, self._deflate_bufs, self._num_threads)
        bytes_deflated = min(len(data), bgu.block_data_inflated_size * len(deflated_sizes))
        return bytes_deflated, [memoryview(buf)[:sz] for buf, sz in zip(self._deflate_bufs, deflated_sizes)]



import os
import glob
import platform
import subprocess
from setuptools import setup, find_packages
from distutils.extension import Extension


install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]

if "Darwin" == platform.system():
    extra_compile_args = ["-O3", "-Xpreprocessor", "-fopenmp"]
    extra_link_args = ["-lomp"]
else:
    extra_compile_args = ["-O3", "-fopenmp"]
    extra_link_args = ["-fopenmp"]

file_ext = "pyx" if os.environ.get("BUILD_WITH_CYTHON") else "c"
extensions = [
    Extension(
        name="bgzip.bgzip_utils",
        sources=[f"bgzip_utils/bgzip_utils.{file_ext}"],
        libraries=["z"],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

if os.environ.get("BUILD_WITH_CYTHON"):
    from Cython.Build import cythonize
    extensions = cythonize(extensions)

with open("README.md") as fh:
    long_description = fh.read()

def get_version():
    filepath = os.path.join(os.path.dirname(__file__), "bgzip", "version.py")
    if os.path.isfile(filepath):
        # Pick up the version packaged into a wheel or source distribution
        with open(filepath) as fh:
            version = dict()
            exec(fh.read().strip(), version)
            return version['__version__']
    else:
        # Pick up version from git annotated tag (canonical version)
        p = subprocess.run(["git", "describe", "--tags", "--match", "v*.*.*"], stdout=subprocess.PIPE)
        p.check_returncode()
        out = p.stdout.decode("ascii").strip()
        if "-" in out:
            out = out.split("-", 1)[0]
        assert out.startswith("v")
        return out[1:]

setup(
    name='bgzip',
    version=get_version(),
    description='Utilities working with blocked gzip streams.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DataBiosphere/bgzip.git',
    author='Brian Hannafious',
    author_email='bhannafi@ucsc.edu',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    ext_modules=extensions,
    scripts=glob.glob('scripts/*'),
    zip_safe=False,
    install_requires=install_requires,
    platforms=['Posix', 'MacOS X'],
    test_suite='test',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7'
    ]
)



from cpython.object cimport PyObject, Py_buffer

cdef extern from "Python.h":
    extern PyObject * PyByteArray_FromStringAndSize(char *, int) nogil
    extern char * PyByteArray_AS_STRING(PyObject *) nogil
    extern int PyByteArray_Resize(PyObject *, int) nogil
    extern int  PyByteArray_Size(PyObject *) nogil

    extern PyObject * PyList_New(int size) nogil
    extern int PyList_SetItem(PyObject *, int, PyObject *) nogil
    extern PyObject * PyList_GetItem(PyObject *, int) nogil

    extern int PyBytes_GET_SIZE(PyObject *) nogil
    extern char * PyBytes_AS_STRING(PyObject *) nogil

    extern int PyMemoryView_Check(PyObject *) nogil
    extern char * PyMemoryView_GET_BUFFER(PyObject *) nogil
    extern PyObject * PyMemoryView_GET_BASE(PyObject *) nogil

    extern int PySequence_Size(PyObject *) nogil

    extern int PyObject_GetBuffer(PyObject *, Py_buffer *, int flags)
    extern void PyBuffer_Release(Py_buffer *)
    extern int PyBUF_SIMPLE

    extern void Py_INCREF(PyObject *) nogil
    extern void Py_DECREF(PyObject *) nogil
